from qt import *
from qtcanvas import *
import speakercode

__all__ = ["TranscriptEditSideBarCanvas","TranscriptEditSideBarBox"]

class TranscriptEditSideBarBox(QCanvasRectangle):
    def __init__(self,seg,field,format,x,y,w,h,canvas):
        """
        @param seg: Segment (a row of the transcript model)
        @param field: The field to be displayed
        @param format: Display format, e.g. "%s"
        @param x: X-coordinate
        @param y: Y-coordinate
        @param w: Width
        @param h: Height
        @param canvas: The canvas on with this box will be drawn.
        """
        QCanvasRectangle.__init__(self, x, y, w, h, canvas)
        self.seg = seg  # a row of the table model
        self.field = field
        self.format = format

    def drawShape(self, p):
        """
        Called by Qt when this box is drawn.
        
        @param p: An instance of QPainter
        """
        QCanvasRectangle.drawShape(self,p)
        spkr = self.format % self.seg[self.field]
	if spkr is None:
	    spkr = 'None'
        p.font().setPixelSize(8)
        p.drawText(self.rect(),Qt.AlignAuto, spkr)

class TranscriptEditSideBarCanvas(QCanvasView):
    def __init__(self, te, field, format,
                 colorMap=lambda x:Qt.white, parent=None):
        """
        @param te: TranscriptEdit
        @param field: Field to be displayed
        @param format: Field display format
        @param colorMap: speaker color map
        """
        QCanvasView.__init__(self, parent)
        self._canvas = QCanvas(self)
        self._te = te
        self._boxes = {}
        self._colorMap = colorMap
        self._field = field
        self._format = format
        self._data = None
        
        self.setCanvas(self._canvas)
        self.setFrameShape(QFrame.NoFrame)
        self.setVScrollBarMode(self.AlwaysOff)
        self.setHScrollBarMode(self.AlwaysOff)
        self._canvas.setBackgroundColor(Qt.lightGray)

        self.connect(self._te, PYSIGNAL("sideBarLayoutChanged"), self.repaint)

    def resizeEvent(self, e):
        s0 = e.oldSize()
        s1 = e.size()
        self._canvas.resize(s1.width(), s1.height())

    def setData(self, data):
        self._data = data
        self.repaint(self._te.getSideBarLayout())
        self.connect(data.emitter,PYSIGNAL("cellChanged"),self._cellChanged)
            
    def repaint(self, layout):
        self._boxes = {}
        for item in self._canvas.allItems():
            item.setCanvas(None)
            del item

        w = self.width()
        for para,seg,y,h in layout:
            item = TranscriptEditSideBarBox(
                seg, self._field, self._format, 0, y, w, h, self._canvas)
            self._boxes[para] = item
            color = QColor(self._colorMap(seg[self._field]))
            item.setBrush(QBrush(color))
            item.show()
        self._canvas.update()

    # model-to-self
    def _cellChanged(self, i, c, v, w):
        """
        @param i: row index
        @param c: column index
        @param v: cell value
        @param w: old value
        """
        h = self._data.getColumnName(c)
        if h == self._field:
            p = self._te.getParagraphIndex(i)
            if p is not None and p in self._boxes:
                item = self._boxes[p]
                seg = self._data[i]
                color = QColor(self._colorMap(v))
                item.setBrush(QBrush(color))
                item.update()
                self._canvas.update()
                
